from ..identification import IdentificationTechnique
from .processing import mask_to_proto, detection_double_thresh

import operator
import numpy as np
import xarray as xr


class DoubleThresholdIdentification(IdentificationTechnique):

    def __init__(self, field, outer_threshold, inner_threshold, comparison_operator,
                 processing_mode="2d", compress=True, **kwargs):
        self.field = field
        self.outer_threshold = outer_threshold
        self.inner_threshold = inner_threshold
        # Save comparison operator name and get associated comparison function
        self.comparison_operator = comparison_operator
        if comparison_operator in { ">=", "ge" }:
            self.compare = operator.ge
        elif comparison_operator in { ">", "gt" }:
            self.compare = operator.gt
        elif comparison_operator in { "<=", "le" }:
            self.compare = operator.le
        elif comparison_operator in { "<", "lt" }:
            self.compare = operator.lt
        else:
            raise ValueError(f"unknown comparison operator '{comparison_operator}'")
        # ...
        assert self.compare(self.inner_threshold, self.outer_threshold)
        # Compression of object mask array in protobuf objects
        self.compress = compress
        # Specify processing mode 2d or 3d: In 2d, identification will be performed on 2d levels,
        #   in 3d the identification will be performed per 3d block
        assert processing_mode in { "2d", "3d" }
        self.processing_mode = processing_mode

    def precompute(self, dataset: xr.Dataset, **kwargs):
        assert self.field in dataset
        return dataset

    def identify(self, dataset: xr.Dataset, **kwargs):
        field = dataset[self.field].values

        if self.processing_mode == "2d":
            field = field.reshape((1,) + field.shape)

        # TODO give operator to routine
        ids, outer_features, inner_features = detection_double_thresh(
            feature_id=0,
            data=field,
            lower_threshold=self.outer_threshold,
            threshold=self.inner_threshold,
            entire_globe=True
        )

        # Let's say you detected 5 objects:
        obj_list = []

        for id_ in ids:
            # Feature 0 is the "no-feature feature"
            if id_ == 0: continue

            # get an instance of a new object, can pass an ID or set in manually afterwards
            obj = self.get_new_object()
            # set some ID to it
            obj.id = id_ # TODO Needs offset? What happens when executed in parallel?

            # get properties of object and populate them (like defined in template.proto)
            properties = obj.properties
            properties.outer_threshold = self.outer_threshold
            properties.inner_threshold = self.inner_threshold
            properties.comparison_operator = ">" # TODO

            mask = np.zeros_like(field, dtype=np.int8)
            mask[outer_features == id_] = 1
            mask[inner_features == id_] = 2
            if self.processing_mode == "2d":
                mask = mask.reshape(mask.shape[1:])

            mask_to_proto(mask, properties.mask, compress=self.compress)

            obj_list.append(obj)

        # return the dataset (can be changed here), and the list of objects
        return dataset, obj_list

    def postprocess(self, dataset: xr.Dataset, pb2_desc, **kwargs):
        # Here post process: called once after processing the whole dataset.
        # If needed, the dataset or objects can be changed here.

        # objects are already in finished protobuf description, cf. the JSON output.
        # pb2_desc.sets is list of trackable sets, each one contains the timesteps list, where each entry is a valid time and contains a list of objects.
        return dataset, pb2_desc
